"""A two-dimensional vector."""
from typing import Any, Generic, TypeVar

from seligimus.python.decorators.operators.equality.equal_instance_attributes import \
    equal_instance_attributes
from seligimus.python.decorators.operators.equality.equal_type import equal_type

T = TypeVar('T', int, float, complex)  # pylint: disable=invalid-name


class Vector2(Generic[T]):
    """A two-dimensional vector."""
    def __init__(self, x: T, y: T):  # pylint: disable=invalid-name
        self.x: T = x  # pylint: disable=invalid-name
        self.y: T = y  # pylint: disable=invalid-name

    @equal_type
    @equal_instance_attributes
    def __eq__(self, other: Any) -> bool:
        return True

    def __bool__(self) -> bool:
        return bool(self.x) or bool(self.y)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(x={self.x}, y={self.y})'

    def __add__(self, other_position: 'Vector2') -> 'Vector2':
        x: T = self.x + other_position.x  # pylint: disable=invalid-name
        y: T = self.y + other_position.y  # pylint: disable=invalid-name
        sum_ = Vector2(x, y)
        return sum_



"""A decorator for equality operators to check that the other object has the same type."""
from functools import wraps
from typing import Any, Callable


def equal_type(equality_operator: Callable[[Any, Any], bool]) -> Callable[[Any, Any], bool]:
    """Return an equality operator which first checks if the other object has the same type."""
    @wraps(equality_operator)
    def equality_operator_with_equal_type(self: Any, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return equality_operator(self, other)

    return equality_operator_with_equal_type



"""A decorator for equality operators to check that the other object has equal instance
attributes."""
from functools import wraps
from typing import Any, Callable, Dict

from seligimus.python.classes.attributes import get_instance_attributes


def equal_instance_attributes(
        equality_operator: Callable[[Any, Any], bool]) -> Callable[[Any, Any], bool]:
    """Return an equality operator which first checks if the other object has equal instance
    attributes."""
    @wraps(equality_operator)
    def equality_operator_with_equal_instance_attributes(self: Any, other: Any) -> bool:
        self_attributes: Dict[str, Any] = get_instance_attributes(self)
        other_attributes: Dict[str, Any] = get_instance_attributes(other)

        if not self_attributes == other_attributes:
            return False

        return equality_operator(self, other)

    return equality_operator_with_equal_instance_attributes



"""A size in the two-dimensional integer lattice."""
from seligimus.maths.vector_2 import Vector2


class IntegerSize2D(Vector2[int]):
    """A size in the two-dimensional integer lattice."""
    def __init__(self, width: int, height: int):  # pylint: disable=useless-super-delegation
        super().__init__(width, height)

    @property
    def width(self) -> int:
        """Return the width."""
        return self.x

    @width.setter
    def width(self, width: int) -> None:
        self.x = width  # pylint: disable=invalid-name

    @property
    def height(self) -> int:
        """Return the height."""
        return self.y

    @height.setter
    def height(self, height: int) -> None:
        self.y = height  # pylint: disable=invalid-name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(width={self.width}, height={self.height})'



"""Setters and getters for Python class attributes."""
from typing import Any, Dict


def get_instance_attributes(instance: Any) -> Dict[str, Any]:
    """Return the instance attributes of an instance."""
    attributes: Dict[str, Any] = instance.__dict__
    return attributes


def set_attributes(instance: Any, attributes: Dict[str, Any]) -> None:
    """Set the instance attributes of an instance."""
    for attribute_name, attribute_value in attributes.items():
        setattr(instance, attribute_name, attribute_value)



"""A position in the two-dimensional integer lattice."""
from seligimus.maths.vector_2 import Vector2


class IntegerPosition2D(Vector2[int]):  # pylint: disable=too-few-public-methods
    """A position in the two-dimensional integer lattice."""



"""Check if a type is a subclass of a generic."""
from typing import Type


def is_subclass_of_generic(class_: Type, generic_class: Type) -> bool:
    """Return whether or not a class is a subclass of a generic class."""
    original_bases = class_.__orig_bases__

    # yapf: disable
    return any(
        original_base == generic_class
        and original_base.__args__ == generic_class.__args__
        for original_base in original_bases
    )
    # yapf: enable



"""A build script using setuptools for the seligimus package."""
import pathlib
from typing import List

import setuptools  # type: ignore

_HERE = pathlib.Path(__file__).parent

_README = _HERE / 'README.md'
_LONG_DESCRIPTION = _README.read_text()

_VERSION_FILE = _HERE / 'VERSION.txt'
_VERSION = _VERSION_FILE.read_text()

_PACKAGES: List[str] = setuptools.find_packages(where=_HERE, include=['seligimus*'])

setuptools.setup(
    name='seligimus',
    version=_VERSION,
    author='Austin Scola',
    author_email='austinscola@gmail.com',
    description='A mono-library.',
    long_description=_LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/pypa/sampleproject',
    packages=_PACKAGES,
    package_data={'seligimus': ['py.typed']},
    zip_safe=False,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
        'Typing :: Typed',
    ],
    python_requires='>=3.6',
)
